from __future__ import annotations

import os
import re
import secrets
from collections import namedtuple
from contextlib import closing
from io import TextIOWrapper
from pathlib import Path
from typing import Callable, Generator, Iterator, Sequence, Tuple

import pexpect
import pytest
from faker import Faker

from ai.backend.plugin.entrypoint import find_build_root
from ai.backend.test.utils.cli import EOF, ClientRunnerFunc, run

_rx_env_export = re.compile(r"^(export )?(?P<key>\w+)=(?P<val>.*)$")

User = namedtuple(
    "User",
    (
        "username",
        "full_name",
        "email",
        "password",
        "role",
        "status",
        "domain_name",
        "need_password_change",
    ),
)
KeypairOption = namedtuple(
    "KeypairOption", ("is_active", "is_admin", "rate_limit", "resource_policy")
)


@pytest.fixture(scope="session")
def client_bin() -> Path:
    return find_build_root(Path(__file__)) / "backend.ai"


def make_run_fixture(profile_env: str) -> Callable[[Path], Iterator[ClientRunnerFunc]]:
    @pytest.fixture(scope="session")
    def run_given_profile(
        client_bin: Path,
    ) -> Iterator[ClientRunnerFunc]:
        env_from_file = get_env_from_profile(profile_env)

        def run_impl(cmdargs: Sequence[str | Path], *args, **kwargs) -> pexpect.spawn:
            return run([client_bin, *cmdargs], *args, **kwargs, env={**os.environ, **env_from_file})

        yield run_impl

    return run_given_profile


def get_env_from_profile(profile_env: str) -> dict[str, str]:
    if not (file_path_env := os.environ.get(profile_env)):
        raise RuntimeError(f"Missing {profile_env} enviroment variable!")
    file_path = Path(file_path_env)

    envs = {}
    lines = file_path.read_text().splitlines()
    for line in lines:
        if m := _rx_env_export.search(line.strip()):
            envs[m.group("key")] = m.group("val")
    return envs


@pytest.fixture
def domain_name() -> str:
    return f"testing-{secrets.token_hex(8)}"


@pytest.fixture
def temp_domain(domain_name: str, run: ClientRunnerFunc) -> Iterator[str]:
    run(["admin", "domains", "add", domain_name])
    print("==== temp_domain created ====")
    try:
        yield domain_name
    finally:
        with closing(run(["admin", "domains", "purge", domain_name])) as p:
            p.expect_exact("Are you sure?")
            p.sendline("Y")
            p.expect(EOF)


@pytest.fixture(scope="module")
def users(n: int = 3) -> Tuple[User, ...]:
    fake = Faker()
    return tuple(
        User(
            username=fake.user_name(),
            full_name=fake.name(),
            email=fake.email(),
            password=fake.password(8),
            role=["user", "admin", "monitor"][i % 3],
            status=["active", "inactive", "before-verification", "deleted"][i % 4],
            domain_name="default",
            need_password_change=[True, False, True][i % 3],
        )
        for i in range(3)
    )


@pytest.fixture
def gen_username() -> Callable[[], str]:
    return lambda: Faker().user_name()


@pytest.fixture
def gen_fullname() -> Callable[[], str]:
    return lambda: Faker().name()


@pytest.fixture(scope="module")
def keypair_options() -> Tuple[KeypairOption, ...]:
    return (
        KeypairOption(is_active=False, is_admin=True, rate_limit=25000, resource_policy="default"),
        KeypairOption(is_active=True, is_admin=False, rate_limit=None, resource_policy="default"),
        KeypairOption(is_active=True, is_admin=True, rate_limit=30000, resource_policy="default"),
    )


@pytest.fixture(scope="module")
def new_keypair_options() -> Tuple[KeypairOption, ...]:
    return (
        KeypairOption(is_active=True, is_admin=False, rate_limit=15000, resource_policy="default"),
        KeypairOption(is_active=False, is_admin=True, rate_limit=15000, resource_policy="default"),
        KeypairOption(
            is_active=False, is_admin=False, rate_limit=100000, resource_policy="default"
        ),
    )


@pytest.fixture(scope="module")
def keypair_resource_policy() -> str:
    fake = Faker()
    return fake.unique.word()


@pytest.fixture
def txt_file() -> Generator[TextIOWrapper, None, None]:
    filepath = "test.txt"
    with open(filepath, "w") as f:
        f.write("This file is for testing.")
    yield f

    os.remove(filepath)


run_user = make_run_fixture("user_file")
run_user2 = make_run_fixture("user2_file")
run_admin = make_run_fixture("admin_file")



import json
from contextlib import closing
from typing import Callable, Tuple

from ...utils.cli import EOF, ClientRunnerFunc, decode
from ..conftest import User


def test_add_user(run_admin: ClientRunnerFunc, users: Tuple[User, ...]):
    """
    Testcase for user addition.
    """
    print("[ Add user ]")

    # Add users
    for i, user in enumerate(users):
        add_arguments = [
            "--output=json",
            "admin",
            "user",
            "add",
            "-u",
            user.username,
            "-n",
            user.full_name,
            "-r",
            user.role,
            "-s",
            user.status,
            "default",
            user.email,
            user.password,
        ]
        if user.need_password_change:
            add_arguments.append("--need-password-change")
        with closing(run_admin(add_arguments)) as p:
            p.expect(EOF)
            response = json.loads(decode(p.before))
            assert response.get("ok") is True, f"Account creation failed: Account#{i + 1}"

    # Check if user is added
    with closing(run_admin(["--output=json", "admin", "user", "list"])) as p:
        p.expect(EOF)
        decoded = decode(p.before)
        loaded = json.loads(decoded)
        user_list = loaded.get("items")

    assert isinstance(user_list, list), "Expected user list"
    added_users = tuple(get_user_from_list(user_list, user.username) for user in users)

    for i, (added_user, user) in enumerate(zip(added_users, users)):
        assert bool(added_user), f"Added account doesn't exist: Account#{i + 1}"
        assert added_user.get("email") == user.email, f"E-mail mismatch: Account#{i + 1}"
        assert added_user.get("full_name") == user.full_name, f"Full name mismatch: Account#{i + 1}"
        assert added_user.get("status") == user.status, f"User status mismatch: Account#{i + 1}"
        assert added_user.get("role") == user.role, f"Role mismatch: Account#{i + 1}"
        assert added_user.get("need_password_change") is user.need_password_change, (
            f"Password change status mismatch: Account#{i + 1}"
        )


def test_update_user(
    run_admin: ClientRunnerFunc,
    users: Tuple[User, ...],
    gen_username: Callable[[], str],
    gen_fullname: Callable[[], str],
):
    """
    Run this testcase after test_add_user.
    Testcase for user update.
    TODO: User update with roles is not fully covered yet.
    """
    print("[ Update user ]")

    # updated_users = [user.copy() for user in users]
    updated_users = (
        User(
            username=gen_username(),
            full_name=gen_fullname(),
            email=user.email,
            password=user.password,
            role=["user", "admin", "monitor"][i % 3],
            status=["inactive", "active", "active"][i % 3],
            domain_name="default",
            need_password_change=[False, True, False][i % 3],
        )
        for i, user in enumerate(users)
    )

    # Update user
    for updated_user, user in zip(updated_users, users):
        update_arguments = [
            "--output=json",
            "admin",
            "user",
            "update",
            "-u",
            updated_user.username,
            "-n",
            updated_user.full_name,
            "-s",
            updated_user.status,
            "-r",
            updated_user.role,
            "-d",
            updated_user.domain_name,
            user.email,
        ]
        if updated_user.need_password_change:
            update_arguments.append("--need-password-change")
        with closing(run_admin(update_arguments)) as p:
            p.expect(EOF)

    # Check if user is updated correctly
    with closing(run_admin(["--output=json", "admin", "user", "list"])) as p:
        p.expect(EOF)
        after_update_decoded = decode(p.before)
        after_update_loaded = json.loads(after_update_decoded)
        updated_user_list = after_update_loaded.get("items")
        assert isinstance(updated_user_list, list), "Expected user list"

    for i, updated_user in enumerate(updated_users):
        user_dict: dict = get_user_from_list(updated_user_list, updated_user.username)
        assert bool(user_dict), f"Account not found - Account#{i + 1}"
        assert user_dict.get("full_name") == updated_user.full_name, (
            f"Full name mismatch: Account#{i + 1}"
        )
        assert user_dict.get("status") == updated_user.status, (
            f"User status mismatch: Account#{i + 1}"
        )
        assert user_dict.get("role") == updated_user.role, f"Role mismatch: Account#{i + 1}"
        assert user_dict.get("need_password_change") is updated_user.need_password_change, (
            f"Password change status mismatch: Account#{i + 1}"
        )
        assert user_dict.get("domain_name") == updated_user.domain_name, (
            f"Domain mismatch: Account#{i + 1}"
        )


def test_delete_user(run_admin: ClientRunnerFunc, users: Tuple[User, ...]):
    """
    !!Run this testcase after running test_add_user
    Testcase for user deletion.
    """
    print("[ Delete user ]")

    for i, fake_user in enumerate(users):
        with closing(run_admin(["--output=json", "admin", "user", "purge", fake_user.email])) as p:
            p.sendline("y")
            p.expect(EOF)
            before = decode(p.before)
            response = json.loads(before[before.index("{") :])
            assert response.get("ok") is True, f"Account deletion failed: Account#{i + 1}"


def test_list_user(run_admin: ClientRunnerFunc):
    """
    Testcase for user listing.
    """
    with closing(run_admin(["--output=json", "admin", "user", "list"])) as p:
        p.expect(EOF)
        decoded = decode(p.before)
        loaded = json.loads(decoded)
        user_list = loaded.get("items")
        assert isinstance(user_list, list)


def get_user_from_list(users: list, username: str) -> dict:
    for user in users:
        if user.get("username") == username:
            return user
    return {}



import asyncio
from abc import abstractmethod
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager as actxmgr
from typing import Any, Optional, override
from uuid import UUID

from ai.backend.client.session import AsyncSession
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.group import CreatedGroupContext
from ai.backend.test.contexts.image import ImageContext
from ai.backend.test.contexts.session import (
    BootstrapScriptContext,
    ClusterContext,
    CreatedSessionMetaContext,
    SessionContext,
)
from ai.backend.test.contexts.sse import (
    SSEContext,
)
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.data.session import CreatedSessionMeta
from ai.backend.test.templates.session.utils import (
    verify_session_events,
)
from ai.backend.test.templates.template import (
    WrapperTestTemplate,
)


class _BaseInteractiveSessionTemplate(WrapperTestTemplate):
    def _build_session_params(
        self,
        session_name: str,
    ) -> dict[str, Any]:
        image_dep = ImageContext.current()
        cluster_dep = ClusterContext.current()
        session_dep = SessionContext.current()

        params: dict[str, Any] = {
            "image": image_dep.name,
            "resources": session_dep.resources,
            "type_": "interactive",
            "name": session_name,
            "cluster_mode": cluster_dep.cluster_mode,
            "cluster_size": cluster_dep.cluster_size,
        }
        params.update(self._extra_session_params())
        return params

    @abstractmethod
    def _extra_session_params(self) -> dict[str, Any]:
        raise NotImplementedError("Subclasses must implement the _extra_session_params method.")

    async def _verify_session_creation(
        self, client_session: AsyncSession, session_name: str
    ) -> UUID:
        timeout = SSEContext.current().timeout

        listener = asyncio.create_task(
            asyncio.wait_for(
                verify_session_events(
                    client_session,
                    session_name,
                    "session_started",
                    {"session_failure", "session_cancelled"},
                ),
                timeout,
            )
        )

        created = await client_session.ComputeSession.get_or_create(
            **self._build_session_params(session_name),
        )

        assert created.created, "Session creation failed"
        assert created.name == session_name
        assert created.status == "RUNNING", f"Expected RUNNING, got {created.status}"
        if created.id is None:
            raise RuntimeError("Session ID is None after creation")
        await listener
        return created.id

    async def _verify_session_destruction(
        self, client_session: AsyncSession, session_name: str
    ) -> None:
        timeout = SSEContext.current().timeout

        listener = asyncio.create_task(
            asyncio.wait_for(
                verify_session_events(
                    client_session,
                    session_name,
                    "session_terminated",
                    {"session_failure", "session_cancelled"},
                    expected_termination_reason="user-requested",
                ),
                timeout,
            )
        )

        result = await client_session.ComputeSession(session_name).destroy()
        assert result["stats"]["status"] == "terminated", (
            f"Expected terminated, got {result['stats']}"
        )
        await listener

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        spec_meta = TestSpecMetaContext.current()
        session_name = f"test_session_{spec_meta.test_id}"
        client_session = ClientSessionContext.current()
        session_id: Optional[UUID] = None

        try:
            session_id = await self._verify_session_creation(client_session, session_name)
            with CreatedSessionMetaContext.with_current(
                CreatedSessionMeta(id=session_id, name=session_name)
            ):
                yield
        finally:
            if session_id:
                await self._verify_session_destruction(client_session, session_name)


class InteractiveSessionTemplate(_BaseInteractiveSessionTemplate):
    @property
    def name(self) -> str:
        return "interactive_session"

    @override
    def _extra_session_params(self) -> dict[str, Any]:
        return {}


class InteractiveSessionWithBootstrapScriptTemplate(_BaseInteractiveSessionTemplate):
    @property
    def name(self) -> str:
        return "interactive_session_with_bootstrap_script"

    @override
    def _extra_session_params(self) -> dict[str, Any]:
        script_ctx = BootstrapScriptContext.current()
        return {"bootstrap_script": script_ctx.bootstrap_script}


class InteractiveSessionWithCustomGroupTemplate(_BaseInteractiveSessionTemplate):
    @property
    def name(self) -> str:
        return "interactive_session_with_custom_group"

    @override
    def _extra_session_params(self) -> dict[str, Any]:
        created_group_meta = CreatedGroupContext.current()
        print(f"Creating endpoint for group: {created_group_meta.group_name}")
        return {"group_name": created_group_meta.group_name}



from typing import override

from ai.backend.client.exceptions import BackendAPIError
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.group import CreatedGroupContext
from ai.backend.test.templates.template import TestCode


class PurgeGroupFailRemainingActiveEndpointExist(TestCode):
    @override
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        created_group_meta = CreatedGroupContext.current()

        try:
            await client_session.Group.purge(str(created_group_meta.group_id))
        except BackendAPIError as e:
            assert e.status == 400, (
                "Expected a 400 Bad Request error when purging a group with active endpoints."
            )


class PurgeGroupFailRemainingActiveSessionExist(TestCode):
    @override
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        created_group_meta = CreatedGroupContext.current()

        try:
            await client_session.Group.purge(str(created_group_meta.group_id))
        except BackendAPIError as e:
            assert e.status == 400, (
                "Expected a 400 Bad Request error when purging a group with active sessions."
            )



import asyncio
from typing import override
from uuid import UUID

from ai.backend.client.session import AsyncSession
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.model_service import (
    CreatedModelServiceEndpointMetaContext,
    ModelServiceContext,
)
from ai.backend.test.templates.model_service.utils import wait_until_all_inference_sessions_ready
from ai.backend.test.templates.template import TestCode

_ENDPOINT_CREATION_TIMEOUT = 30


class ReplicasScaleSuccessfully(TestCode):
    """
    Test case to verify that the number of replicas can be successfully modified.
    """

    @override
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        model_service_dep = ModelServiceContext.current()
        original_replicas_count = model_service_dep.replicas
        model_service_endpoint_meta = CreatedModelServiceEndpointMetaContext.current()
        service_id = model_service_endpoint_meta.service_id

        vfolder_id = await client_session.VFolder(
            name=model_service_dep.model_vfolder_name
        ).get_id()

        # Verify scaling up the replicas
        scaling_up_replicas = original_replicas_count + 2
        await client_session.Service(id=service_id).scale(scaling_up_replicas)
        await self._verify_replicas_count(
            client_session=client_session,
            endpoint_id=service_id,
            expected_replicas=scaling_up_replicas,
            vfolder_id=vfolder_id,
        )

        # Verify scaling down the replicas
        scaling_down_replicas = scaling_up_replicas - 1
        await client_session.Service(id=service_id).scale(scaling_down_replicas)
        await self._verify_replicas_count(
            client_session=client_session,
            endpoint_id=service_id,
            expected_replicas=scaling_down_replicas,
            vfolder_id=vfolder_id,
        )

    async def _verify_replicas_count(
        self,
        client_session: AsyncSession,
        endpoint_id: UUID,
        expected_replicas: int,
        vfolder_id: UUID,
    ) -> None:
        await asyncio.wait_for(
            wait_until_all_inference_sessions_ready(
                client_session=client_session,
                endpoint_id=endpoint_id,
                replicas=expected_replicas,
                vfolder_id=vfolder_id,
            ),
            timeout=_ENDPOINT_CREATION_TIMEOUT,
        )
        info = await client_session.Service(id=endpoint_id).info()
        assert info["replicas"] == expected_replicas, (
            f"Expected replicas count {expected_replicas}, but got {info['replicas']}."
        )
        assert info["desired_session_count"] == expected_replicas, (
            f"Expected desired session count {expected_replicas}, "
            f"but got {info['desired_session_count']}."
        )



from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.vfolder import CreatedVFolderMetaContext, UploadedFilesContext
from ai.backend.test.templates.template import TestCode


class VFolderListFilesSuccess(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        upload_files_ctx = UploadedFilesContext.current()

        response = await client_session.VFolder(vfolder_meta.name).list_files("")

        assert "items" in response, "Response does not contain 'items' key."

        files_in_response = list(map(lambda x: x["name"], response["items"]))

        for file in upload_files_ctx.files:
            assert file.path in files_in_response, f"{file.path} is missing from the response"



class DependencyNotSet(Exception):
    """Raised when a required dependency is not set."""

    def __init__(self, message="Required dependency is not set."):
        super().__init__(message)


class UnexpectedSuccess(Exception):
    """Raised when a test marked as expected to fail actually succeeds."""

    def __init__(self, message="Test succeeded unexpectedly."):
        super().__init__(message)


class UnexpectedFailure(Exception):
    """Raised when a test marked as expected to succeed actually fails."""

    def __init__(self, message="Test failed unexpectedly."):
        super().__init__(message)



from contextlib import asynccontextmanager as actxmgr
from typing import AsyncIterator, override

from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.user import CreatedUserContext
from ai.backend.test.contexts.vfolder import (
    CreatedVFolderMetaContext,
    VFolderContext,
    VFolderShareContext,
)
from ai.backend.test.data.vfolder import VFolderShareMeta
from ai.backend.test.templates.template import WrapperTestTemplate
from ai.backend.test.utils.exceptions import UnexpectedFailure


class ShareVFolderTemplate(WrapperTestTemplate):
    @property
    def name(self) -> str:
        return "share_vfolder"

    @override
    @actxmgr
    async def _context(self) -> AsyncIterator[None]:
        vfolder_meta = CreatedVFolderMetaContext.current()
        client_session = ClientSessionContext.current()
        created_user_meta = CreatedUserContext.current()
        vfolder_dep = VFolderContext.current()
        override_permission = vfolder_dep.share_permission

        vfolder_info = await client_session.VFolder(name=vfolder_meta.name).info()
        # TODO: Move VFolderOwnershipType to common package
        if vfolder_info["type"] != "group":
            raise UnexpectedFailure("VFolder type should be 'group' for sharing.")

        try:
            await client_session.VFolder(name=vfolder_meta.name).share(
                perm=override_permission, emails=[created_user_meta.email]
            )
            with VFolderShareContext.with_current(
                VFolderShareMeta(
                    override_permission=override_permission, user_emails=[created_user_meta.email]
                )
            ):
                yield
        finally:
            await client_session.VFolder(name=vfolder_meta.name).unshare(
                emails=[created_user_meta.email]
            )
            vfolder_info = await client_session.VFolder(name=vfolder_meta.name).info()
            assert vfolder_info["permission"] == vfolder_dep.permission, (
                f"VFolder permission should be reverted to '{vfolder_dep.permission}' after unsharing."
            )



from __future__ import annotations

from pathlib import Path
from typing import Protocol, Sequence

import pexpect

EOF = pexpect.EOF
TIMEOUT = pexpect.TIMEOUT


class ClientRunnerFunc(Protocol):
    def __call__(
        self,
        cmdargs: Sequence[str | Path],
        *args,
        **kwargs,
    ) -> pexpect.spawn:
        pass


def run(
    args: Sequence[str | Path],
    *,
    default_timeout: int = 5,
    **kwargs,
) -> pexpect.spawn:
    p = pexpect.spawn(
        str(args[0]),
        [str(arg) for arg in args[1:]],
        timeout=default_timeout,
        **kwargs,
    )
    return p


def decode(pexpect_capture: bytes | None) -> str:
    if pexpect_capture is None:
        return ""
    return pexpect_capture.decode()



from ai.backend.client.func.session import ComputeSession
from ai.backend.common.types import ClusterMode
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.image import ImageContext
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.templates.template import TestCode


class InteractiveSessionCreationFailureScheduleTimeout(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        image_dep = ImageContext.current()
        spec_meta = TestSpecMetaContext.current()
        test_id = spec_meta.test_id
        session_name = f"test_failure_{str(test_id)}"

        resp: ComputeSession = await client_session.ComputeSession.get_or_create(
            image_dep.name,
            architecture=image_dep.architecture,
            name=session_name,
            enqueue_only=False,
            max_wait=5,
            resources={
                "cpu": 9999999999  # Intentionally large value to trigger timeout
            },
            cluster_mode=ClusterMode.SINGLE_NODE,
            cluster_size=1,
        )

        assert resp.status == "TIMEOUT", (
            f"Session should have timed out due to scheduling failure, actual status: {resp.status}"
        )



from typing import override

from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.session import SessionDependencyContext
from ai.backend.test.templates.template import TestCode


class DependencyGraphRetriever(TestCode):
    @override
    async def test(self) -> None:
        session = ClientSessionContext.current()
        session_dependency = SessionDependencyContext.current()

        for dependent, dependencies in session_dependency.dependencies.items():
            session_name = dependent.name
            result = await session.ComputeSession(name=session_name).get_dependency_graph()

            actual_dependencies = result["depends_on"]
            assert len(actual_dependencies) == len(dependencies), (
                f"Expected {len(dependencies)} dependencies, got {len(actual_dependencies)}"
            )

            actual_dependency_ids = {dep["session_id"] for dep in actual_dependencies}
            expected_dependency_ids = {str(dep.id) for dep in dependencies}

            assert actual_dependency_ids == expected_dependency_ids, (
                f"Expected dependencies {expected_dependency_ids}, "
                f"got {actual_dependency_ids} for session {session_name}"
            )



from typing import override

from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.session import CreatedSessionMetaContext
from ai.backend.test.templates.template import TestCode


class TestContainerLogRetriever(TestCode):
    @override
    async def test(self) -> None:
        session = ClientSessionContext.current()
        session_meta = CreatedSessionMetaContext.current()
        session_id = session_meta.id
        session_name = session_meta.name

        session_info = await session.ComputeSession.from_session_id(session_id=session_id).detail()
        kernels_info = session_info["kernels"]
        for kernel in kernels_info:
            result = await session.ComputeSession(name=session_name).get_logs(
                kernel_id=kernel["row_id"]
            )
            assert result["result"]["logs"] is not None
            assert result["result"]["logs"] != ""



import tempfile
from pathlib import Path

from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.vfolder import CreatedVFolderMetaContext, UploadedFilesContext
from ai.backend.test.templates.template import TestCode


class VFolderDownloadSuccess(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        upload_files_ctx = UploadedFilesContext.current()

        with tempfile.TemporaryDirectory() as download_dir:
            download_root = Path(download_dir)

            for uploaded_file in upload_files_ctx.files:
                (download_root / uploaded_file.path).parent.mkdir(parents=True, exist_ok=True)

                await client_session.VFolder(vfolder_meta.name).download(
                    [uploaded_file.path],
                    basedir=download_root,
                    max_retries=3,
                )

                local_path = download_root / uploaded_file.path
                assert local_path.exists(), f"{local_path} is missing"
                assert local_path.read_text() == uploaded_file.content, (
                    f"{local_path} file content mismatch"
                )



from typing import override

from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.session import CreatedSessionMetaContext
from ai.backend.test.templates.template import TestCode


class TestSessionRename(TestCode):
    @override
    async def test(self) -> None:
        session = ClientSessionContext.current()
        session_meta = CreatedSessionMetaContext.current()
        session_id = session_meta.id
        session_name = session_meta.name
        new_name = f"{session_name}_renamed"

        await session.ComputeSession(name=session_name).rename(new_name)
        renamed_info = await session.ComputeSession.from_session_id(session_id=session_id).detail()
        assert renamed_info["name"] == new_name

        # Clean up by renaming back to original name
        await session.ComputeSession(name=new_name).rename(session_name)



import textwrap

from ai.backend.common.types import ClusterMode
from ai.backend.test.contexts.context import ContextName
from ai.backend.test.templates.auth.keypair import KeypairAuthTemplate
from ai.backend.test.templates.model_service.endpoint import (
    EndpointTemplate,
    PublicEndpointTemplate,
)
from ai.backend.test.templates.model_service.jwt_token import ModelServiceTokenTemplate
from ai.backend.test.testcases.model_service.health_check import (
    EndpointHealthCheck,
    EndpointHealthCheckWithToken,
)
from ai.backend.test.testcases.model_service.replicas_scale_successfully import (
    ReplicasScaleSuccessfully,
)
from ai.backend.test.testcases.model_service.scale_failure_too_many_sessions import (
    ReplicasScaleFailureTooManySessions,
)
from ai.backend.test.testcases.spec_manager import TestSpec, TestTag
from ai.backend.test.tester.dependency import ClusterDep

from ...templates.template import BasicTestTemplate

SINGLE_NODE_SINGLE_CONTAINER_MODEL_SERVICE_TEST_SPECS = {
    "single_node_single_container_creation_endpoint_success": TestSpec(
        name="single_node_single_container_creation_endpoint_success",
        description=textwrap.dedent("""\
            Test for successful creation of an endpoint.
            This test verifies that an endpoint can be created successfully.
            The test will:
            1. Create an endpoint with a specified name and resources.
            2. Verify that the endpoint is not available without generating JWT token.
            3. Clean up the endpoint after verification.
        """),
        tags={
            TestTag.MANAGER,
            TestTag.AGENT,
            TestTag.MODEL_SERVICE,
            TestTag.SESSION,
            TestTag.SINGLE_NODE_SINGLE_CONTAINER,
        },
        # Endpoint health check failure is expected.
        template=BasicTestTemplate(
            EndpointHealthCheck(expected_status_codes={400, 401})
        ).with_wrappers(KeypairAuthTemplate, EndpointTemplate),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=1,
                ),
            ]
        },
    ),
    "single_node_single_container_creation_public_endpoint_success": TestSpec(
        name="single_node_single_container_creation_public_endpoint_success",
        description=textwrap.dedent("""\
            Test for successful creation of an endpoint.
            This test verifies that an endpoint can be created successfully.
            The test will:
            1. Create an endpoint with a specified name and resources.
            2. Verify that the endpoint is created and available.
            3. Clean up the endpoint after verification.
        """),
        tags={
            TestTag.MANAGER,
            TestTag.AGENT,
            TestTag.MODEL_SERVICE,
            TestTag.SESSION,
            TestTag.SINGLE_NODE_SINGLE_CONTAINER,
        },
        template=BasicTestTemplate(EndpointHealthCheck(expected_status_codes={200})).with_wrappers(
            KeypairAuthTemplate, PublicEndpointTemplate
        ),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=1,
                ),
            ]
        },
    ),
    "single_node_single_container_creation_private_endpoint_success": TestSpec(
        name="single_node_single_container_creation_private_endpoint_success",
        description=textwrap.dedent("""\
            Test for successful creation of an endpoint.
            This test verifies that an endpoint can be created successfully.
            The test will:
            1. Create an endpoint with a specified name and resources.
            2. Verify that the endpoint is created and available.
            3. Clean up the endpoint after verification.
        """),
        tags={
            TestTag.MANAGER,
            TestTag.AGENT,
            TestTag.MODEL_SERVICE,
            TestTag.SESSION,
            TestTag.SINGLE_NODE_SINGLE_CONTAINER,
        },
        # Endpoint health check success is expected since we're inject JWT token into the request.
        template=BasicTestTemplate(
            EndpointHealthCheckWithToken(expected_status_codes={200})
        ).with_wrappers(KeypairAuthTemplate, EndpointTemplate, ModelServiceTokenTemplate),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=1,
                ),
            ]
        },
    ),
    "replicas_successfully_scale": TestSpec(
        name="replicas_successfully_scale",
        description=textwrap.dedent("""\
            Test for successful scaling of replicas in an endpoint.
            This test verifies that the number of replicas can be successfully scaled.
            The test will:
            1. Create an endpoint with a specified number of replicas.
            2. Scale up the replicas and verify the change.
            3. Scale down the replicas and verify the change.
        """),
        tags={TestTag.MANAGER, TestTag.AGENT, TestTag.MODEL_SERVICE, TestTag.SESSION},
        template=BasicTestTemplate(ReplicasScaleSuccessfully()).with_wrappers(
            KeypairAuthTemplate, EndpointTemplate
        ),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=1,
                ),
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=3,
                ),
                ClusterDep(
                    cluster_mode=ClusterMode.MULTI_NODE,
                    cluster_size=3,
                ),
            ]
        },
    ),
    "replicas_scale_failure_too_many_sessions": TestSpec(
        name="replicas_scale_failure_too_many_sessions",
        description=textwrap.dedent("""\
            Test for failure when trying to scale replicas beyond the allowed limit.
            This test verifies that scaling replicas beyond the allowed limit raises an error.
            The test will:
            1. Create an endpoint with a specified number of replicas.
            2. Attempt to scale up the replicas beyond the allowed limit set in user resource policy and verify the error.
        """),
        tags={TestTag.MANAGER, TestTag.AGENT, TestTag.MODEL_SERVICE, TestTag.SESSION},
        template=BasicTestTemplate(ReplicasScaleFailureTooManySessions()).with_wrappers(
            KeypairAuthTemplate, EndpointTemplate
        ),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=1,
                ),
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=3,
                ),
                ClusterDep(
                    cluster_mode=ClusterMode.MULTI_NODE,
                    cluster_size=3,
                ),
            ]
        },
    ),
}

SINGLE_NODE_MULTI_CONTAINER_MODEL_SERVICE_TEST_SPECS = {
    "single_node_multi_container_creation_endpoint_success": TestSpec(
        name="single_node_multi_container_creation_endpoint_success",
        description=textwrap.dedent("""\
            Test for successful creation of an endpoint.
            This test verifies that an endpoint can be created successfully.
            The test will:
            1. Create an endpoint with a specified name and resources.
            2. Verify that the endpoint is not available without generating JWT token.
            3. Clean up the endpoint after verification.
        """),
        tags={
            TestTag.MANAGER,
            TestTag.AGENT,
            TestTag.MODEL_SERVICE,
            TestTag.SESSION,
            TestTag.REQUIRED_SINGLE_NODE_MULTI_CONTAINER_CONFIGURATION,
        },
        # Endpoint health check failure is expected.
        template=BasicTestTemplate(
            EndpointHealthCheck(expected_status_codes={400, 401})
        ).with_wrappers(KeypairAuthTemplate, EndpointTemplate),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=3,
                ),
            ]
        },
    ),
    "single_node_multi_container_creation_public_endpoint_success": TestSpec(
        name="single_node_multi_container_creation_public_endpoint_success",
        description=textwrap.dedent("""\
            Test for successful creation of an endpoint.
            This test verifies that an endpoint can be created successfully.
            The test will:
            1. Create an endpoint with a specified name and resources.
            2. Verify that the endpoint is created and available.
            3. Clean up the endpoint after verification.
        """),
        tags={
            TestTag.MANAGER,
            TestTag.AGENT,
            TestTag.MODEL_SERVICE,
            TestTag.SESSION,
            TestTag.REQUIRED_SINGLE_NODE_MULTI_CONTAINER_CONFIGURATION,
        },
        template=BasicTestTemplate(EndpointHealthCheck(expected_status_codes={200})).with_wrappers(
            KeypairAuthTemplate, PublicEndpointTemplate
        ),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=3,
                ),
            ]
        },
    ),
    "single_node_multi_container_creation_private_endpoint_success": TestSpec(
        name="single_node_multi_container_creation_private_endpoint_success",
        description=textwrap.dedent("""\
            Test for successful creation of an endpoint.
            This test verifies that an endpoint can be created successfully.
            The test will:
            1. Create an endpoint with a specified name and resources.
            2. Verify that the endpoint is created and available.
            3. Clean up the endpoint after verification.
        """),
        tags={
            TestTag.MANAGER,
            TestTag.AGENT,
            TestTag.MODEL_SERVICE,
            TestTag.SESSION,
            TestTag.REQUIRED_SINGLE_NODE_MULTI_CONTAINER_CONFIGURATION,
        },
        # Endpoint health check success is expected since we're inject JWT token into the request.
        template=BasicTestTemplate(
            EndpointHealthCheckWithToken(expected_status_codes={200})
        ).with_wrappers(KeypairAuthTemplate, EndpointTemplate, ModelServiceTokenTemplate),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.SINGLE_NODE,
                    cluster_size=3,
                ),
            ]
        },
    ),
}

MULTI_NODE_MULTI_CONTAINER_MODEL_SERVICE_TEST_SPECS = {
    "multi_node_multi_container_creation_endpoint_success": TestSpec(
        name="multi_node_multi_container_creation_endpoint_success",
        description=textwrap.dedent("""\
            Test for successful creation of an endpoint.
            This test verifies that an endpoint can be created successfully.
            The test will:
            1. Create an endpoint with a specified name and resources.
            2. Verify that the endpoint is not available without generating JWT token.
            3. Clean up the endpoint after verification.
        """),
        tags={
            TestTag.MANAGER,
            TestTag.AGENT,
            TestTag.MODEL_SERVICE,
            TestTag.SESSION,
            TestTag.REQUIRED_MULTI_NODE_MULTI_CONTAINER_CONFIGURATION,
        },
        # Endpoint health check failure is expected.
        template=BasicTestTemplate(
            EndpointHealthCheck(expected_status_codes={400, 401})
        ).with_wrappers(KeypairAuthTemplate, EndpointTemplate),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.MULTI_NODE,
                    cluster_size=3,
                ),
            ]
        },
    ),
    "multi_node_multi_container_creation_public_endpoint_success": TestSpec(
        name="multi_node_multi_container_creation_public_endpoint_success",
        description=textwrap.dedent("""\
            Test for successful creation of an endpoint.
            This test verifies that an endpoint can be created successfully.
            The test will:
            1. Create an endpoint with a specified name and resources.
            2. Verify that the endpoint is created and available.
            3. Clean up the endpoint after verification.
        """),
        tags={
            TestTag.MANAGER,
            TestTag.AGENT,
            TestTag.MODEL_SERVICE,
            TestTag.SESSION,
            TestTag.REQUIRED_MULTI_NODE_MULTI_CONTAINER_CONFIGURATION,
        },
        template=BasicTestTemplate(EndpointHealthCheck(expected_status_codes={200})).with_wrappers(
            KeypairAuthTemplate, PublicEndpointTemplate
        ),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.MULTI_NODE,
                    cluster_size=3,
                ),
            ]
        },
    ),
    "multi_node_multi_container_creation_private_endpoint_success": TestSpec(
        name="multi_node_multi_container_creation_private_endpoint_success",
        description=textwrap.dedent("""\
            Test for successful creation of an endpoint.
            This test verifies that an endpoint can be created successfully.
            The test will:
            1. Create an endpoint with a specified name and resources.
            2. Verify that the endpoint is created and available.
            3. Clean up the endpoint after verification.
        """),
        tags={
            TestTag.MANAGER,
            TestTag.AGENT,
            TestTag.MODEL_SERVICE,
            TestTag.SESSION,
            TestTag.REQUIRED_MULTI_NODE_MULTI_CONTAINER_CONFIGURATION,
        },
        # Endpoint health check success is expected since we're inject JWT token into the request.
        template=BasicTestTemplate(
            EndpointHealthCheckWithToken(expected_status_codes={200})
        ).with_wrappers(KeypairAuthTemplate, EndpointTemplate, ModelServiceTokenTemplate),
        parametrizes={
            ContextName.CLUSTER_CONFIG: [
                ClusterDep(
                    cluster_mode=ClusterMode.MULTI_NODE,
                    cluster_size=3,
                ),
            ]
        },
    ),
}

MODEL_SERVICE_TEST_SPECS = {
    **SINGLE_NODE_SINGLE_CONTAINER_MODEL_SERVICE_TEST_SPECS,
    **SINGLE_NODE_MULTI_CONTAINER_MODEL_SERVICE_TEST_SPECS,
    **MULTI_NODE_MULTI_CONTAINER_MODEL_SERVICE_TEST_SPECS,
}



import enum
from abc import abstractmethod
from collections.abc import Iterator, Mapping, MutableMapping
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Generic, Optional, TypeVar, final

T = TypeVar("T")


class ContextName(enum.StrEnum):
    """
    Enum for context names used in the test framework.
    This enum provides a set of predefined context names that can be used to identify different contexts in the testing framework.
    """

    TEST_SPEC_META = "test_spec_meta"
    IMAGE = "image"
    CLIENT_SESSION = "client_session"
    ENDPOINT = "endpoint"
    KEYPAIR = "keypair"
    LOGIN_CREDENTIAL = "login_credential"
    SESSION = "session"
    SSE = "sse"
    DOMAIN = "domain"
    GROUP = "group"
    SCALING_GROUP = "scaling_group"
    USER_RESOURCE_POLICY = "user_resource_policy"
    MODEL_SERVICE = "model_service"
    BATCH_SESSION = "batch_session"
    CLUSTER_CONFIG = "cluster_config"
    CODE_EXECUTION = "code_execution"
    SESSION_TEMPLATE = "session_template"
    BOOTSTRAP_SCRIPT = "bootstrap_script"
    SESSION_IMAGIFY = "session_imagify"
    TEST_ERROR_OUTPUT_DIRECTORY = "test_error_output_directory"
    SESSION_DEPENDENCY = "session_dependency"

    VFOLDER = "vfolder"
    VFOLDER_UPLOAD_FILES = "vfolder_upload_files"
    VFOLDER_SHARE_META = "vfolder_share_meta"
    VFOLDER_INVITATION = "vfolder_invitation"
    VFOLDER_INVITATION_PERMISSION = "vfolder_invitation_permission"
    VFOLDER_UPLOADED_FILES_META = "vfolder_uploaded_files_meta"

    CREATED_VFOLDER_META = "created_vfolder_meta"
    CREATED_SESSION_META = "created_session_meta"
    CREATED_SESSION_TEMPLATE_ID = "created_session_template_id"
    CREATED_MODEL_SERVICE_ENDPOINT = "created_model_service_endpoint"
    CREATED_MODEL_SERVICE_TOKEN = "created_model_service_token"
    CREATED_USER_CONTEXT = "created_user_context"
    CREATED_USER_CLIENT_SESSION = "created_user_client_session"
    CREATED_GROUP = "created_group"


class BaseTestContext(Generic[T]):
    """
    Base class for tester's context management.
    This class provides a way to manage context variables for test scenarios.
    It uses `contextvars` to store the current value and provides methods to get and set this value.
    It is designed to be subclassed for specific test contexts.
    """

    _ctxvar: Optional[ContextVar[Optional[T]]] = None
    _used: MutableMapping[ContextName, "BaseTestContext"] = {}

    def __init_subclass__(cls):
        if cls._ctxvar is not None:
            raise RuntimeError(f"{cls.__name__} is already initialized")
        cls._ctxvar = ContextVar[T](f"{cls.__name__}_ctxvar", default=None)
        cls._used[cls.name()] = cls

    @classmethod
    @abstractmethod
    def name(cls) -> ContextName:
        """
        Get the name of the context
        :return: name of the context
        """
        raise NotImplementedError(
            f"{cls.__name__} must implement get_name method to return the context name"
        )

    @classmethod
    def used_contexts(cls) -> Mapping[ContextName, "BaseTestContext"]:
        """
        Get all used contexts
        :return: mapping of context names to context instances
        """
        return cls._used

    @classmethod
    @final
    def current(cls) -> T:
        """
        Get the current value from context
        :raises RuntimeError: if no value is set in the context (Tester must set it before using)
        """
        if cls._ctxvar is None:
            raise RuntimeError("Don't use BaseTestContext directly, subclass it instead")
        res = cls._ctxvar.get()
        if res is None:
            raise RuntimeError(f'No value is set in "{cls.name()}" context')
        return res

    @classmethod
    @final
    @contextmanager
    def with_current(cls, value: T) -> Iterator[None]:
        """Set the current value in the context"""
        if cls._ctxvar is None:
            raise RuntimeError("Don't use BaseTestContext directly, subclass it instead")
        token = cls._ctxvar.set(value)
        try:
            yield
        finally:
            cls._ctxvar.reset(token)



from ai.backend.test.contexts.context import BaseTestContext, ContextName
from ai.backend.test.tester.dependency import DomainDep


class DomainContext(BaseTestContext[DomainDep]):
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.DOMAIN
